function [theta_out, Hval, DHval, D2Hval, iter] = ...
                     thetafun_sim(theta_in, WfdCell, U, scrrng, ...
                                  Slambda, logdensfd, ...
                                  itermax, crit)
%  Computes scores minimizing the regularized negative log likelihood
%  This verson has explicit arguments rather than using a struct object
%  to input data and logdensfd.  It is intended for simulation purposes
%  so that parfor loops can be used.
%  
%  Arguments are:
%  THETA_IN ... Initial values for abilities in [0,n].  It is assumed
%               that values corresponding to rows of U that are all
%               0's will be 0, and values corresponding to all 1's will be
%               n's.
%               This is a column vector of size N, where N is the number
%               of examinees.  The numbers will be in [0,n] and will be
%               floating point numbers rather than integers, where n is
%               the number of test items.
%  WFDCELL  ... A cell array of length n.  This is used in functions
%               Hfun and DHfun to define the item type, either binary or 
%               multi-option, and the number of options if the item type
%               is multi-option.
%  U        ... An N by n matrix of indicating which option was
%               selected for each examinee and each item.
%  SCRRNG   ... Range of admissible score values.  Defaults to [0,n],
%               but for 3PL W functions, for example, this is input
%               as [0.1, n-0.1]
%  SLAMBDA  ... A smoothing parameter used to shrink ability parameters in
%               THETA towards the corresponding quantile of the 
%               scaled beta distribution fit to the sum scores.
%               This is a non-negative float number.
%  LOGDENSFD... A functional data object used to represent  
%               the log-density of a representation of the prior
%               distribution of the distribution of ability.
%               Not used if Slambda = 0, and if not supplied, is empty;
%  ITERMAX  ... Maximum number of iterations for computing the optimal
%               theta values.  Default is 20.
%               This is a nonnegative integer.
%  CRIT     ... Criterion for convergence of optimization.  Default is 1e-8
%               This is a positive float.
%
%  Last modified 8 May 2018 by Jim Ramsay

%  set default argument values

if nargin < 3
    error('Less than three arguments supplied.');
end

[N,n] = size(U);
Uactmat = ones(N,n,'logical');

if nargin < 8,  crit      = 1e-8;   end 
if nargin < 7,  itermax   = 20;     end 
if nargin < 6,  logdensfd = [];     end
if nargin < 5,  Slambda   =  0;     end 
if nargin < 4,  scrrng    = [0,n];  end 

%  check logdensfd

if Slambda > 0
    if ~isa_fd(logdensfd)
        error('Object logdensfd is not a functional data object.');
    end
end

%  indices of scores on the boundaries and in the interior

scrmin  = scrrng(1);
scrmax  = scrrng(2);

theta_in(theta_in <= scrmin) = scrmin;
theta_in(theta_in >= scrmax) = scrmax;

%  Initial function, first and second derivative values wrt theta

Hval             =  Hfun(theta_in, WfdCell, U, Uactmat);
[DHval, D2Hval]  = DHfun(theta_in, WfdCell, U, Uactmat);

%  add penalty if required

if Slambda > 0
    tQval   = eval_fd(theta_in, tQ);
    DtQval  = eval_fd(theta_in, tQ, 1);
    D2tQval = eval_fd(theta_in, tQ, 2);
    Hval   = Hval   - Slambda.*tQval;
    DHval  = DHval  - Slambda.*DtQval;
    D2Hval = D2Hval - Slambda.*D2tQval;
end

if itermax == 0 
    theta_out = theta_in;
    return;  
end

%  Initialize indices of active theta values

% active = (theta_in  > scrmin & theta_in < scrmax) | ...
%          (theta_in == scrmin & DHval < 0) | ...
%          (theta_in == scrmax & DHval > 0);
% nactive   = sum(active);

active   = ActiveTestFn(theta_in, DHval, D2Hval, scrmin, scrmax, crit);
nactive  = sum(active);
theta_ina = theta_in(active);

%  find cases where gradient size exceeds criterion and further
%  optimization is required

%  update scores in active set by a Newton-Raphson step

iter = 0;
% Ha   = Hval(active);
DHa  = DHval(active);
D2Ha = D2Hval(active);
while nactive > 0 && iter < itermax
    iter = iter + 1;
%     disp(['iter = ',num2str(iter), ',  nactive = ',num2str(nactive)])
%     disp([theta_in(active), Ha, DHa, D2Ha])
    %  compute Newton-Raphson step sizes
    step = DHa./D2Ha;
    % disp(['untested step = ',num2str(step')])
    %  bound the steps between -2 and 2
    step(step < -2) = -2; 
    step(step >  2) =  2; 
    % disp(['2 bounded step = ',num2str(step')])
    %  ensure that initial step does not go to boundaries
    stepind0 = find(theta_ina - step <= scrmin);
    stepindn = find(theta_ina - step >= scrmax);
    if ~isempty(stepind0)  
        step(stepind0) =  (theta_ina(stepind0) - scrmin);  
    end
    if ~isempty(stepindn) 
        step(stepindn) = -(scrmax - theta_ina(stepindn));
    end
    % disp(['score bounded step = ',num2str(step')])
    %  half fac as required to achieve the reduction of all fn. values
    %  step through the iteration while any(Hval(active) - Ha < -1e-2)
    %  up to a maximum number of 10 step halvings
    theta_inanew  = theta_in(active);
    Uanew         = U(active,:);
    Uactmatanew   = Uactmat(active,:);
    Hanew         = Hval(active);
    fac      = 1;
    stepiter = 0;
    %  initial set of theta's failing to meet criterion
    failindex  = 1:nactive;
    while ~isempty(failindex) && stepiter < 10
        stepiter = stepiter + 1;
        %  compute current step size
        stepnew     = fac*step(failindex);
        theta_inanewfail = theta_inanew(failindex);
        stepnewind0 = find(theta_inanewfail - stepnew <= scrmin);
        stepnewindn = find(theta_inanewfail - stepnew >= scrmax);
        if ~isempty(stepnewind0)  
            stepnew(stepnewind0) = ...
                 (theta_inanewfail(stepnewind0) - scrmin);  
        end
        if ~isempty(stepnewindn)  
            stepnew(stepnewindn) = ...
                -(scrmax - theta_inanewfail(stepnewindn));  
        end
        %  update ability values
        theta_inanew(failindex) = theta_inanew(failindex) - stepnew;
%         if any(theta_inanew < scrrng(1)) || any(theta_inanew > scrrng(2))
%             disp([min(theta_inanew),max(theta_inanew)])
%             error('Values in theta_inanew out of range,');
%         end
        %  Compute new function value
%         disp(['size(active): ',num2str(size(active))])
%         disp(['size(failindex): ',num2str(size(failindex))])
%         disp(['size(theta_inanew(failindex)): ', ...
%               num2str(size(theta_inanew(failindex)))])
%         disp(['size(Hanew): ',num2str(size(Hanew))])        
%         disp(['size(active(failindex)): ', ...
%               num2str(size(active(failindex)))]) 
%         disp(['failindex: ',num2str(failindex)])
%         disp(['active(failindex): ', ...
%               num2str(active(failindex)')]) 
        Hanew(failindex) = Hfun(theta_inanew(failindex),   WfdCell, ...
                                       Uanew(failindex,:), ...
                                 Uactmatanew(failindex,:)); 
        if Slambda > 0
            if isempty(theta_inanew(failindex))
                error('theta_inanew(failindex) is empty.');
            end
            tQval = eval_fd(theta_inanew(failindex), tQ);
            Hanew(failindex) = Hanew(failindex) - Slambda.*tQval;
        end
        failindex = find(Hval(active) - Hanew < -crit);
        %  halve fac in preparation for next step size iteration
        fac = fac/2;
    end
    %  either all function values reduced or 10 halvings completed
    %  save current function values
    theta_in(active) = theta_inanew;
    Ha = Hanew;
    [DHa, D2Ha] = ...
        DHfun(theta_in(active), WfdCell, U(active,:), Uactmat(active,:));
    if Slambda > 0
        DtQval  = eval_fd(theta_inanew, tQ, 1);
        D2tQval = eval_fd(theta_inanew, tQ, 2);
        DHa  = DHa  - Slambda.*DtQval;
        D2Ha = D2Ha - Slambda.*D2tQval;
    end
    Hval(active)   = Ha;
    DHval(active)  = DHa;
    D2Hval(active) = D2Ha;
    %  find theta values that need further iterations
%     % 1:  abs(slope) > crit and D2H positive 
%     test1 = abs(DHa) > crit & D2Ha > 0 & ...
%             theta_inanew > scrmin & theta_inanew < scrmax;
%     % 2:  theta at scrmin and slope positive
%     test2 = theta_inanew == scrmin & DHa < 0;
%     % 3:  theta at scrmax and slope negative
%     test3 = theta_inanew == scrmax & DHa > 0;
%     active  = active(test1 | test2 | test3);
    active = ActiveTestFn(theta_inanew, DHa, D2Ha, scrmin, ...
                          scrmax, crit, active);
    nactive = sum(active);
    % disp(['nactive = ',num2str(nactive)])
    if nactive > 0
        theta_ina = theta_in(active);
        DHa  = DHval(active);
        D2Ha = D2Hval(active);
    end
    %  return for a new optimization step 
end

%  opimization complete for all cases, save results and exit

theta_out = theta_in;



